#!/usr/bin/env python3
from urllib.parse import urlparse

import argparse
import sys
import logging
from typing import Any, Iterator, Optional

from .challenge_parser import parse_challenge, TARGET_INFO_NAMES
from .requester import request_http, request_SMBv1, request_SMBv23

from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from threading import Thread
import json

logger = logging.getLogger(__name__)

DONE = -1


def parse_args():
    parser = argparse.ArgumentParser(
      description='Fetch and parse NTLM challenge messages from HTTP and SMB'
      ' services'
    )
    parser.add_argument(
        'target',
        help='HTTP or SMB URL to fetch NTLM challenge from',
        nargs="*",
    )
    parser.add_argument(
        '-1', '--smbv1',
        action='store_true',
        help='Use SMBv1 (for SMB requests)'
    )

    parser.add_argument(
        "-v", "--verbose",
        action="count",
        help="Verbosity",
        default=0
    )
    parser.add_argument(
        "-w", "--workers",
        help="Set the number of workers",
        default=10,
        type=int
    )

    parser.add_argument(
        "-j", "--json",
        metavar="FILE",
        help="Write json results in given file",
        type=argparse.FileType('w'),
    )

    return parser.parse_args()


def main():
    args = parse_args()
    init_log(args.verbose)

    pool = ThreadPoolExecutor(args.workers)
    q = Queue()
    t_print = launch_printer(q, args.json)

    for url in read_text_targets(args.target):
        pool.submit(work, q, url, args.smbv1)

    pool.shutdown(wait=True)
    q.put(DONE)
    t_print.join()


def init_log(verbosity=0, log_file=None):

    if verbosity == 1:
        level = logging.WARN
    elif verbosity == 2:
        level = logging.INFO
    elif verbosity > 2:
        level = logging.DEBUG
    else:
        level = logging.CRITICAL

    logging.basicConfig(
        level=level,
        filename=log_file,
        format="%(levelname)s:%(name)s:%(message)s"
    )


def work(queue, url, smbv1=False):
    try:
        challenge = request_ntlm_challenge(url, smbv1=smbv1)
        queue.put((url, challenge))
    except Exception as ex:
        logger.warning("Error with '{}': {}".format(url, ex))


def launch_printer(q: Queue, json_file=None):
    results_iter = queue_to_iter(q)
    t_print = Thread(
        target=process_challenges,
        args=(
            results_iter,
        ),
        kwargs={
            "json_file": json_file
        }
    )
    t_print.start()
    return t_print


def queue_to_iter(q: Queue):
    """Allows to process the consumption of the queue as an iterator
    """
    while True:
        value = q.get()
        if value == DONE:
            return
        yield value


def process_challenges(challenges, json_file=None):

    results = []

    for url, challenge in challenges:
        try:
            parsed_challenge = parse_challenge(challenge)
            print_challenge(url, parsed_challenge)
            results.append((url, parsed_challenge))
        except Exception:
            logger.exception("Error parsing challenge of %s", url)

    if json_file and results:
        save_json_results(json_file, results)


DEFAULT_PORTS = {
    "smb": 445,
    "http": 80,
    "https": 443
}


def save_json_results(json_file, results):
    json_array = []
    for url, challenge in results:
        url_p = urlparse(url)

        port = url_p.port
        if port is None:
            port = DEFAULT_PORTS[url_p.scheme]

        json_array.append({
            "target": {
                "url": url,
                "protocol": url_p.scheme,
                "host": url_p.hostname,
                "port": port
            },
            "challenge": {
                "target_name": challenge.target_name,
                "os": {
                    "major_version": challenge.version.major,
                    "minor_version": challenge.version.minor,
                    "build": challenge.version.build,
                    "names": challenge.version.names,
                },
                "target_info": challenge.target_info,
                "negotiate_flags": {
                    "value": challenge.negotiate_flags,
                    "names": challenge.negotiate_flags_names
                },
                "server_challenge": challenge.server_challenge
            }
        })

    logger.info("Results saved in {}".format(json_file.name))
    json.dump(json_array, json_file)


def print_challenge(url, challenge):
    msg = []
    msg.append("\nUrl: {}".format(url))
    if 'NTLMSSP_TARGET_TYPE_DOMAIN' in challenge.negotiate_flags_names:
        msg.append('Target (Domain): {}'.format(challenge.target_name))
    elif 'NTLMSSP_TARGET_TYPE_SERVER' in challenge.negotiate_flags_names:
        msg.append('Target (Server): {}'.format(challenge.target_name))

    version = challenge.version
    msg.append("OS Version: {}.{}.{}".format(
        version.major, version.minor, version.build
    ))
    msg.append("OS Name: {}".format(" | ".join(version.names)))

    target_info = challenge.target_info
    for name in TARGET_INFO_NAMES:
        try:
            msg.append("{}: {}".format(name, target_info[name]))
        except KeyError:
            pass

    msg.append('Negotiate Flags: {}'.format(hex(challenge.negotiate_flags)))
    for name in challenge.negotiate_flags_names:
        msg.append("  {}".format(name))

    msg.append("Server challenge: {}".format(challenge.server_challenge))

    print("\n".join(msg), flush=True)


def request_ntlm_challenge(url, smbv1=False):
    url_p = urlparse(url)

    if url_p.scheme == 'smb':
        host = url_p.hostname
        port = url_p.port or 445

        if smbv1:
            return request_SMBv1(host, port)
        else:
            return request_SMBv23(host, port)

    elif url_p.scheme.startswith('http'):
        return request_http(url)

    raise Exception('Invalid URL, expecting http[s]://... or smb://...'.format(url))


def get_default_port(protocol):
    return {
        "smb": 445,
        "http": 80,
        "https": 443
    }[protocol]


def read_text_targets(targets: Any) -> Iterator[str]:
    yield from read_text_lines(read_targets(targets))


def read_targets(targets: Optional[Any]) -> Iterator[str]:
    """Function to process the program ouput that allows to read an array
    of strings or lines of a file in a standard way. In case nothing is
    provided, input will be taken from stdin.
    """
    if not targets:
        yield from sys.stdin

    for target in targets:
        try:
            with open(target) as fi:
                yield from fi
        except FileNotFoundError:
            yield target


def read_text_lines(fd: Iterator[str]) -> Iterator[str]:
    """To read lines from a file and skip empty lines or those commented
    (starting by #)
    """
    for line in fd:
        line = line.strip()
        if line == "":
            continue
        if line.startswith("#"):
            continue

        yield line


if __name__ == '__main__':
    main()



import datetime
from collections import OrderedDict
from typing import List, Optional, Dict

# parsing from "NT LAN Manager (NTLM) Authentication Protocol" v20190923, revision  31.0
# https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/MS-NLMP/%5bMS-NLMP%5d.pdf

TARGET_INFO_NAMES = [
    "MsvAvNbComputerName",
    "MsvAvNbDomainName",
    "MsvAvDnsComputerName",
    "MsvAvDnsDomainName",
    "MsvAvDnsTreeName",
    "MsvAvFlags",
    "MsvAvTimestamp",
    "MsvAvSingleHost",
    "MsvAvTargetName",
    "MsvAvChannelBindings"
]

NTLM_NEGOTIATE_FLAGS = OrderedDict()
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_UNICODE'] = 0x00000001
NTLM_NEGOTIATE_FLAGS['NTLM_NEGOTIATE_OEM'] = 0x00000002
NTLM_NEGOTIATE_FLAGS['NTLMSSP_REQUEST_TARGET'] = 0x00000004
NTLM_NEGOTIATE_FLAGS['UNUSED_10'] = 0x00000008
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_SIGN'] = 0x00000010
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_SEAL'] = 0x00000020
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_DATAGRAM'] = 0x00000040
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_LM_KEY'] = 0x00000080
NTLM_NEGOTIATE_FLAGS['UNUSED_9'] = 0x00000100
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_NTLM'] = 0x00000200
NTLM_NEGOTIATE_FLAGS['UNUSED_8'] = 0x00000400
NTLM_NEGOTIATE_FLAGS['NTLMSSP_ANONYMOUS'] = 0x00000800
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED'] = 0x00001000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_OEM_WORKSTATION_SUPPLIED'] = 0x00002000
NTLM_NEGOTIATE_FLAGS['UNUSED_7'] = 0x00004000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_ALWAYS_SIGN'] = 0x00008000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_TARGET_TYPE_DOMAIN'] = 0x00010000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_TARGET_TYPE_SERVER'] = 0x00020000
NTLM_NEGOTIATE_FLAGS['UNUSED_6'] = 0x00040000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_EXTENDED_SESSIONSECURITY'] = 0x00080000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_IDENTIFY'] = 0x00100000
NTLM_NEGOTIATE_FLAGS['UNUSED_5'] = 0x00200000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_REQUEST_NON_NT_SESSION_KEY'] = 0x00400000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_TARGET_INFO'] = 0x00800000
NTLM_NEGOTIATE_FLAGS['UNUSED_4'] = 0x01000000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_VERSION'] = 0x02000000
NTLM_NEGOTIATE_FLAGS['UNUSED_3'] = 0x10000000
NTLM_NEGOTIATE_FLAGS['UNUSED_2'] = 0x08000000
NTLM_NEGOTIATE_FLAGS['UNUSED_1'] = 0x04000000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_128'] = 0x20000000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_KEY_EXCH'] = 0x40000000
NTLM_NEGOTIATE_FLAGS['NTLMSSP_NEGOTIATE_56'] = 0x80000000


class OSVersion:

    def __init__(
            self,
            major: int, minor: int, build: int,
            names: Optional[List[str]] = None
    ):
        self.major = major
        self.minor = minor
        self.build = build
        self.names = names or []


class NtlmChallenge:

    def __init__(
            self,
            target_name: str,
            version: OSVersion,
            target_info: Dict[str, str],
            negotiate_flags: int,
            server_challenge: str
    ):
        self.target_name = target_name
        self.version = version
        self.target_info = target_info
        self.negotiate_flags = negotiate_flags
        self.server_challenge = server_challenge

    @property
    def negotiate_flags_names(self):
        names = []

        for name, value in NTLM_NEGOTIATE_FLAGS.items():
            if self.negotiate_flags & value:
                names.append(name)

        return names


def parse_challenge(challenge_message):

    # Signature
    # b'NTLMSSP\x00' --> NTLMSSP
    signature = decode_string(challenge_message[0:7])

    # MessageType
    # b'\x02\x00\x00\x00' --> 2
    message_type = decode_int(challenge_message[8:12])

    # TargetNameFields
    target_name_fields = challenge_message[12:20]
    target_name_len = decode_int(target_name_fields[0:2])
    target_name_max_len = decode_int(target_name_fields[2:4])
    target_name_offset = decode_int(target_name_fields[4:8])

    # NegotiateFlags
    negotiate_flags_int = decode_int(challenge_message[20:24])

    # negotiate_flags = parse_negotiate_flags(negotiate_flags_int)

    # ServerChallenge
    server_challenge = challenge_message[24:32]

    # Reserved
    reserved = challenge_message[32:40]

    # TargetInfoFields
    target_info_fields = challenge_message[40:48]
    target_info_len = decode_int(target_info_fields[0:2])
    target_info_max_len = decode_int(target_info_fields[2:4])
    target_info_offset = decode_int(target_info_fields[4:8])

    # Version
    version_bytes = challenge_message[48:56]
    version = parse_version(version_bytes)

    # TargetName
    target_name = challenge_message[target_name_offset:
                                    target_name_offset+target_name_len]
    target_name = decode_string(target_name)

    # TargetInfo
    target_info_bytes = challenge_message[target_info_offset:
                                          target_info_offset+target_info_len]

    target_info = parse_target_info(target_info_bytes)

    return NtlmChallenge(
        target_name,
        version,
        target_info,
        negotiate_flags_int,
        to_hex_string(server_challenge)
    )


def parse_negotiate_flags(negotiate_flags_int):
    negotiate_flags = []

    for name, value in NTLM_NEGOTIATE_FLAGS.items():
        if negotiate_flags_int & value:
            negotiate_flags.append(name)

    return negotiate_flags


def parse_version(version_bytes: bytes) -> OSVersion:

    major_version = version_bytes[0]
    minor_version = version_bytes[1]
    product_build = decode_int(version_bytes[2:4])

    names = []

    if major_version == 5 and minor_version == 1:
        names.append('Windows XP (SP2)')
    elif major_version == 5 and minor_version == 2:
        names.append('Server 2003')
    elif major_version == 6 and minor_version == 0:
        names.append("Server 2008")
        names.append("Windows Vista")
    elif major_version == 6 and minor_version == 1:
        names.append("Server 2008 R2")
        names.append("Windows 7")
    elif major_version == 6 and minor_version == 2:
        names.append("Server 2012")
        names.append("Windows 8")
    elif major_version == 6 and minor_version == 3:
        names.append("Server 2012 R2")
        names.append("Windows 8.1")
    elif major_version == 10 and minor_version == 0:
        names.append("Server 2016")
        names.append("Server 2019")
        names.append("Windows 10")

    return OSVersion(major_version, minor_version, product_build, names)


def parse_target_info(target_info_bytes):

    MsvAvEOL = 0x0000
    MsvAvNbComputerName = 0x0001
    MsvAvNbDomainName = 0x0002
    MsvAvDnsComputerName = 0x0003
    MsvAvDnsDomainName = 0x0004
    MsvAvDnsTreeName = 0x0005
    MsvAvFlags = 0x0006
    MsvAvTimestamp = 0x0007
    MsvAvSingleHost = 0x0008
    MsvAvTargetName = 0x0009
    MsvAvChannelBindings = 0x000A

    target_info = OrderedDict()
    info_offset = 0

    while info_offset < len(target_info_bytes):
        av_id = decode_int(target_info_bytes[info_offset:info_offset+2])
        av_len = decode_int(target_info_bytes[info_offset+2:info_offset+4])
        av_value = target_info_bytes[info_offset+4:info_offset+4+av_len]

        info_offset = info_offset + 4 + av_len

        if av_id == MsvAvEOL:
            pass
        elif av_id == MsvAvNbComputerName:
            target_info["MsvAvNbComputerName"] = decode_string(av_value)
        elif av_id == MsvAvNbDomainName:
            target_info["MsvAvNbDomainName"] = decode_string(av_value)
        elif av_id == MsvAvDnsComputerName:
            target_info["MsvAvDnsComputerName"] = decode_string(av_value)
        elif av_id == MsvAvDnsDomainName:
            target_info["MsvAvDnsDomainName"] = decode_string(av_value)
        elif av_id == MsvAvDnsTreeName:
            target_info["MsvAvDnsTreeName"] = decode_string(av_value)
        elif av_id == MsvAvFlags:
            pass
        elif av_id == MsvAvTimestamp:
            filetime = decode_int(av_value)
            microseconds = (filetime - 116444736000000000) / 10
            time = datetime.datetime(
                1970, 1, 1) + datetime.timedelta(microseconds=microseconds)
            target_info["MsvAvTimestamp"] = time.strftime(
                "%b %d, %Y %H:%M:%S.%f")
        elif av_id == MsvAvSingleHost:
            target_info["MsvAvSingleHost"] = decode_string(av_value)
        elif av_id == MsvAvTargetName:
            target_info["MsvAvTargetName"] = decode_string(av_value)
        elif av_id == MsvAvChannelBindings:
            target_info["MsvAvChannelBindings"] = to_hex_string(av_value)

    return target_info


def to_hex_string(byte_string):
    return ''.join('{:02x}'.format(x) for x in byte_string)


def decode_string(byte_string):
    return byte_string.decode('UTF-8').replace('\x00', '')


def decode_int(byte_string):
    return int.from_bytes(byte_string, 'little')



import base64
import requests
from impacket import smb3, smb, ntlm
import logging

from urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':HIGH:!DH:!aNULL'
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

logger = logging.getLogger(__name__)


def request_http(url):
    logger.debug("Request HTTP for {}".format(url))
    headers = {
        'Authorization': 'NTLM TlRMTVNTUAABAAAAB4IIAAAAAAAAAAAAAAAAAAAAAAA='
    }

    request = requests.get(url, headers=headers, verify=False)

    if request.status_code not in [401, 302]:
        raise Exception(
            'Expecting response code 401 or 302, received: {}'.format(
                request.status_code))

    try:
        auth_header = request.headers['WWW-Authenticate']
    except KeyError:
        raise Exception(
            "NTLM Challenge response not found (WWW-Authenticate header missing)"
        )

    if 'NTLM' not in auth_header:
        raise Exception(
            'NTLM Challenge response not found (WWW-Authenticate does not contain "NTLM")'
        )

    challenge_message = base64.b64decode(
        auth_header.split(' ')[1].replace(',', ''))

    return challenge_message


def request_SMBv23(host, port=445):
    logger.debug("Request SMB2/3 for {}:{}".format(host, port))

    # start client
    smb_client = smb3.SMB3(host, host, sess_port=port)

    # start: modified from login()
    # https://github.com/SecureAuthCorp/impacket/blob/master/impacket/smb3.py

    session_setup = smb3.SMB2SessionSetup()

    if smb_client.RequireMessageSigning is True:
        session_setup['SecurityMode'] = smb3.SMB2_NEGOTIATE_SIGNING_REQUIRED
    else:
        session_setup['SecurityMode'] = smb3.SMB2_NEGOTIATE_SIGNING_ENABLED

    session_setup['Flags'] = 0

    # NTLMSSP
    blob = smb3.SPNEGO_NegTokenInit()
    blob['MechTypes'] = [
        smb3.TypesMech['NTLMSSP - Microsoft NTLM Security Support Provider']]

    auth = ntlm.getNTLMSSPType1(smb_client._Connection['ClientName'], '',
                                smb_client._Connection['RequireSigning'])
    blob['MechToken'] = auth.getData()

    session_setup['SecurityBufferLength'] = len(blob)
    session_setup['Buffer'] = blob.getData()

    packet = smb_client.SMB_PACKET()
    packet['Command'] = smb3.SMB2_SESSION_SETUP
    packet['Data'] = session_setup

    packet_id = smb_client.sendSMB(packet)

    smb_response = smb_client.recvSMB(packet_id)

    if smb_client._Connection['Dialect'] == smb3.SMB2_DIALECT_311:
        smb_client.__UpdatePreAuthHash(smb_response.rawData)

    # NTLM challenge
    if smb_response.isValidAnswer(smb3.STATUS_MORE_PROCESSING_REQUIRED):
        session_setup_response = smb3.SMB2SessionSetup_Response(
            smb_response['Data'])
        resp_token = smb3.SPNEGO_NegTokenResp(session_setup_response['Buffer'])

        return resp_token['ResponseToken']

    else:
        return None


def request_SMBv1(host, port=445):
    logger.debug("Request SMB1 for {}:{}".format(host, port))

    # start client
    smb_client = smb.SMB(host, host, sess_port=port)

    # start: modified from login_standard()
    # https://github.com/SecureAuthCorp/impacket/blob/master/impacket/smb.py

    session_setup = smb.SMBCommand(smb.SMB.SMB_COM_SESSION_SETUP_ANDX)
    session_setup['Parameters'] = smb.SMBSessionSetupAndX_Extended_Parameters()
    session_setup['Data'] = smb.SMBSessionSetupAndX_Extended_Data()

    session_setup['Parameters']['MaxBufferSize'] = 61440
    session_setup['Parameters']['MaxMpxCount'] = 2
    session_setup['Parameters']['VcNumber'] = 1
    session_setup['Parameters']['SessionKey'] = 0
    session_setup['Parameters']['Capabilities'] = smb.SMB.CAP_EXTENDED_SECURITY | \
        smb.SMB.CAP_USE_NT_ERRORS |  \
        smb.SMB.CAP_UNICODE | \
        smb.SMB.CAP_LARGE_READX | \
        smb.SMB.CAP_LARGE_WRITEX

    # NTLMSSP
    blob = smb.SPNEGO_NegTokenInit()

    blob['MechTypes'] = [
        smb.TypesMech['NTLMSSP - Microsoft NTLM Security Support Provider']]

    auth = ntlm.getNTLMSSPType1(smb_client.get_client_name(), '',
                                smb_client._SignatureRequired, use_ntlmv2=True)
    blob['MechToken'] = auth.getData()

    session_setup['Parameters']['SecurityBlobLength'] = len(blob)
    session_setup['Parameters'].getData()
    session_setup['Data']['SecurityBlob'] = blob.getData()
    session_setup['Data']['NativeOS'] = 'Unix'
    session_setup['Data']['NativeLanMan'] = 'Samba'

    smb_packet = smb.NewSMBPacket()

    if smb_client._SignatureRequired:
        smb_packet['Flags2'] |= smb_packet.SMB.FLAGS2_SMB_SECURITY_SIGNATURE

    smb_packet.addCommand(session_setup)

    smb_client.sendSMB(smb_packet)

    smb_response = smb_client.recvSMB()

    # NTLM challenge
    if smb_response.isValidAnswer(smb.SMB.SMB_COM_SESSION_SETUP_ANDX):

        session_response = smb.SMBCommand(smb_response['Data'][0])
        session_parameters = smb.SMBSessionSetupAndX_Extended_Response_Parameters(
            session_response['Parameters'])
        session_data = smb.SMBSessionSetupAndX_Extended_Response_Data(
            flags=smb_response['Flags2'])
        session_data['SecurityBlobLength'] = session_parameters['SecurityBlobLength']
        session_data.fromString(session_response['Data'])

        resp_token = smb.SPNEGO_NegTokenResp(session_data['SecurityBlob'])

        return resp_token['ResponseToken']

    else:
        return None



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()


name = "ntlm-info"

setuptools.setup(
    name=name,
    version="0.0.1",
    author="Eloy Perez",
    author_email="zer1t0ps@protonmail.com",
    description="Get system info from NTLM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "ntlm-info = ntlm_info.main:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)
